from dataclasses import dataclass
from typing import (Any, Callable, Dict, List, Optional, Sequence, Set, Type,
                    Union)

from fastapi import Response, params
from fastapi.datastructures import Default, DefaultPlaceholder
from fastapi.responses import JSONResponse
from fastapi.routing import APIRoute
from starlette.routing import Route

SetIntStr = Set[Union[int, str]]
DictIntStrAny = Dict[Union[int, str], Any]


@dataclass
class RouteArgs:
    """The arguments APIRouter.add_api_route takes.

    Just a convenience for type safety and so we can pass all the args needed by the underlying FastAPI route args via
    `**dataclasses.asdict(some_args)`.
    """
    path: str
    response_model: Optional[Type[Any]] = None
    status_code: Optional[int] = None
    tags: Optional[List[str]] = None
    dependencies: Optional[Sequence[params.Depends]] = None
    summary: Optional[str] = None
    description: Optional[str] = None
    response_description: str = "Successful Response"
    responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None
    deprecated: Optional[bool] = None
    methods: Optional[Union[Set[str], List[str]]] = None
    operation_id: Optional[str] = None
    response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None
    response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None
    response_model_by_alias: bool = True
    response_model_exclude_unset: bool = False
    response_model_exclude_defaults: bool = False
    response_model_exclude_none: bool = False
    include_in_schema: bool = True
    response_class: Union[Type[Response], DefaultPlaceholder] = Default(
        JSONResponse
    )
    name: Optional[str] = None
    route_class_override: Optional[Type[APIRoute]] = None
    callbacks: Optional[List[Route]] = None
    openapi_extra: Optional[Dict[str, Any]] = None

    class Config:
        arbitrary_types_allowed = True


@dataclass
class EndpointDefinition:
    """RouteArgs plus the endpoint.

    Endpoint is separate as it has to be bound to self before it can be used.
    """
    endpoint: Callable[..., Any]
    args: RouteArgs




CBV_CLASS_KEY = "__cbv_class__"
CLASS_TYPE = "__cbv_class__"

SIGNATURE_KEY = "__signature__"
API_METHODS = "__api_methods__"
INIT_MODIFIED = "__init_modified__"
GENERIC_ATTRIBUTES = "__generic_attribute__"
TAGGING_KEY = "TAGGING"


from enum import Enum
from typing import (Any, Callable, Dict, List, Optional, Sequence, Type,
                    Union, TypeVar)

from fastapi import Response, params
from fastapi.datastructures import Default
from fastapi.encoders import DictIntStrAny, SetIntStr
from fastapi.responses import JSONResponse
from starlette.routing import BaseRoute

from .route_args import EndpointDefinition, RouteArgs

AnyCallable = TypeVar('AnyCallable', bound=Callable[..., Any])

CONTROLLER_METHOD_KEY = '__controller_method__'


def route(path: str, methods: List[str], **kwargs: Any) -> Callable[[AnyCallable], AnyCallable]:
    """General purpose route definition. Requires you to pass an array of HTTP methods like GET, POST, PUT, etc.

    The remaining kwargs are exactly the same as for FastAPI's decorators like @get, @post, etc.

    Most users will probably want to use the shorter decorators like @get, @post, @put, etc. so they don't have to pass
    the list of methods.
    """

    def marker(method: AnyCallable, type_=None) -> AnyCallable:
        setattr(
            method, '_endpoint',
            EndpointDefinition(endpoint=method, args=RouteArgs(path=path, methods=methods, **kwargs))
        )
        setattr(
            method, CONTROLLER_METHOD_KEY, True
        )
        return method

    return marker


def get(
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        **kwargs: Any) -> Callable[[AnyCallable], AnyCallable]:
    print(path)
    return route(
        path,
        methods=['GET'],
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        dependencies=dependencies,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        operation_id=operation_id,
        response_model_include=response_model_include,
        response_model_exclude=response_model_exclude,
        response_model_by_alias=response_model_by_alias,
        response_model_exclude_unset=response_model_exclude_unset,
        response_model_exclude_defaults=response_model_exclude_defaults,
        response_model_exclude_none=response_model_exclude_none,
        include_in_schema=include_in_schema,
        response_class=response_class,
        name=name,
        callbacks=callbacks,
        openapi_extra=openapi_extra,
        **kwargs, )


def post(
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        **kwargs: Any) -> Callable[[AnyCallable], AnyCallable]:
    return route(
        path,
        methods=['POST'],
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        dependencies=dependencies,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        operation_id=operation_id,
        response_model_include=response_model_include,
        response_model_exclude=response_model_exclude,
        response_model_by_alias=response_model_by_alias,
        response_model_exclude_unset=response_model_exclude_unset,
        response_model_exclude_defaults=response_model_exclude_defaults,
        response_model_exclude_none=response_model_exclude_none,
        include_in_schema=include_in_schema,
        response_class=response_class,
        name=name,
        callbacks=callbacks,
        openapi_extra=openapi_extra,
        **kwargs
    )


def patch(
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        **kwargs: Any) -> Callable[[AnyCallable], AnyCallable]:
    return route(
        path,
        methods=['PATCH'],
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        dependencies=dependencies,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        operation_id=operation_id,
        response_model_include=response_model_include,
        response_model_exclude=response_model_exclude,
        response_model_by_alias=response_model_by_alias,
        response_model_exclude_unset=response_model_exclude_unset,
        response_model_exclude_defaults=response_model_exclude_defaults,
        response_model_exclude_none=response_model_exclude_none,
        include_in_schema=include_in_schema,
        response_class=response_class,
        name=name,
        callbacks=callbacks,
        openapi_extra=openapi_extra,
        **kwargs
    )


def put(
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        **kwargs: Any) -> Callable[[AnyCallable], AnyCallable]:
    return route(
        path,
        methods=['PUT'],
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        dependencies=dependencies,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        operation_id=operation_id,
        response_model_include=response_model_include,
        response_model_exclude=response_model_exclude,
        response_model_by_alias=response_model_by_alias,
        response_model_exclude_unset=response_model_exclude_unset,
        response_model_exclude_defaults=response_model_exclude_defaults,
        response_model_exclude_none=response_model_exclude_none,
        include_in_schema=include_in_schema,
        response_class=response_class,
        name=name,
        callbacks=callbacks,
        openapi_extra=openapi_extra,
        **kwargs
    )


def delete(
        path: str,
        *,
        response_model: Optional[Type[Any]] = None,
        status_code: Optional[int] = None,
        tags: Optional[List[Union[str, Enum]]] = None,
        dependencies: Optional[Sequence[params.Depends]] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        response_description: str = "Successful Response",
        responses: Optional[Dict[Union[int, str], Dict[str, Any]]] = None,
        deprecated: Optional[bool] = None,
        operation_id: Optional[str] = None,
        response_model_include: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_exclude: Optional[Union[SetIntStr, DictIntStrAny]] = None,
        response_model_by_alias: bool = True,
        response_model_exclude_unset: bool = False,
        response_model_exclude_defaults: bool = False,
        response_model_exclude_none: bool = False,
        include_in_schema: bool = True,
        response_class: Type[Response] = Default(JSONResponse),
        name: Optional[str] = None,
        callbacks: Optional[List[BaseRoute]] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        **kwargs: Any) -> Callable[[AnyCallable], AnyCallable]:
    return route(
        path,
        methods=['DELETE'],
        response_model=response_model,
        status_code=status_code,
        tags=tags,
        dependencies=dependencies,
        summary=summary,
        description=description,
        response_description=response_description,
        responses=responses,
        deprecated=deprecated,
        operation_id=operation_id,
        response_model_include=response_model_include,
        response_model_exclude=response_model_exclude,
        response_model_by_alias=response_model_by_alias,
        response_model_exclude_unset=response_model_exclude_unset,
        response_model_exclude_defaults=response_model_exclude_defaults,
        response_model_exclude_none=response_model_exclude_none,
        include_in_schema=include_in_schema,
        response_class=response_class,
        name=name,
        callbacks=callbacks,
        openapi_extra=openapi_extra,
        **kwargs
    )



import copy
import dataclasses
import inspect
import typing
from collections import defaultdict
from typing import Any, Callable, Dict, List, Tuple, Type, TypeVar, cast, get_type_hints

from fastapi import Depends, APIRouter
from pydantic.typing import is_classvar

from class_based_fastapi.decorators import CONTROLLER_METHOD_KEY
from class_based_fastapi.defaults import GENERIC_ATTRIBUTES, CLASS_TYPE, API_METHODS, INIT_MODIFIED, \
    SIGNATURE_KEY, TAGGING_KEY
from class_based_fastapi.route_args import RouteArgs
from class_based_fastapi.templates_formatting import _rest_api_naming
from class_based_fastapi.utilities import deepcopy_func

AnyCallable = TypeVar('AnyCallable', bound=Callable[..., Any])


class RoutableMeta(type):
    """Это метакласс, который формирует методы API, которые были отмечены декоратором маршрута/пути, в значения
    для конструктора маршрутизации добавления конечных точек к своему маршрутизатору из библиотеки FastAPI."""

    __inheritors__ = defaultdict(list)
    __inheritors_set__ = set()

    @staticmethod
    def get_config_endpoints(bases: Tuple[Type[Any]], attrs: Dict[str, Any]) -> Dict[str, Any]:
        """Получение конфигурации всех методов API

        Args:
            bases: Дочерние типы
            attrs: Атрибуты класса

        Returns: Конфигурация всех методов API

        """
        cls_ = attrs.get(CLASS_TYPE)
        if cls_ is None:
            return {}

        new_methods = list(filter(lambda x: hasattr(x[1], CONTROLLER_METHOD_KEY), attrs.items()))
        old_methods = attrs.get(API_METHODS, dict())
        if not len(old_methods) and len(bases):
            old_methods = {}
            for base in bases:
                if not hasattr(base, API_METHODS):
                    continue
                for name_method, args in getattr(base, API_METHODS).items():
                    old_methods[name_method] = copy.deepcopy(args)

        for name_method, endpoint in new_methods:
            old_methods[name_method] = copy.deepcopy(endpoint._endpoint.args)

        return copy.deepcopy(old_methods)

    @staticmethod
    def _init_cbv(cls: Type[type]) -> None:
        """Идемпотентно изменяет предоставленный "cls", выполняя следующие модификации:

        * Функция `__init__` обновлена для установки любых зависимостей с аннотациями к классу в качестве атрибутов экземпляра.
        * Атрибут `__signature__` обновляется, чтобы указать FastAPI, какие аргументы следует передавать инициализатору

        Args:
            cls: Тип модифицируемого класса

        Returns: None
        """
        old_init: Callable[..., Any] = cls.__init__
        if getattr(old_init, INIT_MODIFIED, False):
            return
        old_signature = inspect.signature(old_init)
        old_parameters = list(old_signature.parameters.values())[1:]  # drop `self` parameter
        new_parameters = [
            x for x in old_parameters if x.kind not in (inspect.Parameter.VAR_POSITIONAL, inspect.Parameter.VAR_KEYWORD)
        ]
        dependency_names: List[str] = []
        for name, hint in get_type_hints(cls).items():
            if is_classvar(hint) or name == '_endpoints':
                continue
            parameter_kwargs = {"default": getattr(cls, name, Ellipsis)}
            dependency_names.append(name)
            new_parameters.append(
                inspect.Parameter(name=name, kind=inspect.Parameter.KEYWORD_ONLY, annotation=hint, **parameter_kwargs)
            )
        new_signature = old_signature.replace(parameters=new_parameters)

        def new_init(self: Any, *args: Any, **kwargs: Any) -> None:
            for dep_name in dependency_names:
                try:
                    dep_value = kwargs.pop(dep_name)
                    setattr(self, dep_name, dep_value)
                except Exception as ex:
                    print(ex)
                    pass
            old_init(self, *args, **kwargs)

        setattr(new_init, INIT_MODIFIED, True)
        setattr(cls, "__signature__", new_signature)
        setattr(cls, "__init__", new_init)

        # attrs["__signature__"] = new_signature
        # attrs["__init__"] = new_init
        # attrs[CBV_CLASS_KEY] = True

    @staticmethod
    def __new_instance__(cls: Type[type], name: str, bases: Tuple[Type[Any]], attrs: Dict[str, Any]) -> 'RoutableMeta':
        """Создание нового типа и запоминание родительских типов

        Args:
            cls: Исходный тип (родительский)
            name: Название создаваемого типа
            bases: Родительские типы
            attrs: Атрибуты создаваемого типа

        Returns: Созданный тип
        """
        # Унаследованные классы
        klass = cast(RoutableMeta, type.__new__(cls, name, bases, attrs))
        for base in klass.mro()[1:-1]:
            cls.__inheritors__[base].append(klass)
            cls.__inheritors_set__.add(klass)
        return klass

    @staticmethod
    def get_type_instance(cls: Type[type], name: str):
        """Получение типа, наследующегося от cls

        Args:
            cls: Родительский тип
            name: Название дочернего типа

        Returns: Унаследованный тип, соответствующий названию name

        """
        for class_ in cls.__inheritors_set__:
            if class_.__name__ == name:
                return class_

    @staticmethod
    def _compute_path_new(cls: Type[type], func: Callable) -> Callable:
        """Формирование шаблона URI

       Args:
           cls: Тип класса
           func: Метод API

       Returns: Сформированный метод API
       """

        config_methods = getattr(cls, API_METHODS)
        args = config_methods[func.__name__] if func.__name__ in config_methods else copy.deepcopy(func._endpoint.args)
        func_ = deepcopy_func(func, args, cls)
        user_path = args.path
        base_template = cls.BASE_TEMPLATE_PATH
        if not str.startswith(user_path, '/'):
            user_path = base_template.replace('{user_path}', user_path)

        name_module = cls.NAME_MODULE
        if '{module}' in user_path and (name_module is None or name_module == ''):
            user_path = user_path.replace('/{module}', '')

        # Template
        if str.endswith(user_path, '/'):
            user_path = user_path[:-1]
        args.template_path = user_path

        # Name module
        name_module = _rest_api_naming(name_module)
        args.name_module = name_module

        # Version
        varsion_api = str.lower(cls.VERSION_API)
        args.varsion_api = varsion_api

        # Replacement
        path = user_path \
            .replace('{module}', name_module) \
            .replace('{version}', varsion_api) \
            .replace('{controller}', _rest_api_naming(cls.__name__))
        args.path = path
        # print(f'{args.methods}: {path}')
        return func_

    @staticmethod
    def compute_tags_route(args: RouteArgs, cls: Type[type]):
        if not getattr(cls, TAGGING_KEY, False): return

        args.tags = (args.tags or [])
        if args.tags == []:
            args.tags = (getattr(cls, 'TAGS', None) or [cls.__name__])

    @staticmethod
    def get_router(cls: Type[type]) -> APIRouter:
        """Формирование функции возвращающей маршруты API

        Args:
            cls: Тип класса, содержащий методы API

        Returns: APIRouter
        """
        config_methods = getattr(cls, API_METHODS)
        functions = [
            RoutableMeta._compute_path_new(cls, func)
            for name, func in inspect.getmembers(cls, inspect.isfunction)
            if hasattr(func, CONTROLLER_METHOD_KEY) or name in config_methods
        ]

        router = APIRouter()

        for endpoint in functions:
            # get the signature of the endpoint function
            signature = inspect.signature(endpoint)
            # get the parameters of the endpoint function
            signature_parameters = list(signature.parameters.values())

            # replace the class instance with the itself FastApi Dependecy
            signature_parameters[0] = signature_parameters[0].replace(
                default=Depends(cls)
            )

            # set self and after it the keyword args
            new_parameters = [signature_parameters[0]] + [
                parameter.replace(kind=inspect.Parameter.KEYWORD_ONLY)
                for parameter in signature_parameters[1:]
            ]

            new_signature = signature.replace(parameters=new_parameters)
            setattr(endpoint, SIGNATURE_KEY, new_signature)

            args = config_methods[endpoint.__name__] if endpoint.__name__ in config_methods else endpoint._endpoint.args
            RoutableMeta.compute_tags_route(args, cls)
            router.add_api_route(
                endpoint=endpoint,
                **dataclasses.asdict(args)
            )
        return router

    @staticmethod
    def init_generic_params(cls):
        if not hasattr(cls, '__orig_bases__'):
            return
        bases = cls.__orig_bases__
        generics = list(
            filter(
                lambda x: isinstance(x, typing._GenericAlias) and getattr(x, '__origin__', None) in {typing.Generic},
                bases
            )
        )
        if len(generics) != 0:
            generic = generics[0]

            generic_attrs = list(
                map(
                    lambda x: {
                        'name': x.__name__,
                        'type': None,
                        'class': cls
                    },
                    generic.__args__
                )
            )
            setattr(cls, GENERIC_ATTRIBUTES, generic_attrs)

        generics = list(
            filter(
                lambda x: isinstance(x, typing._GenericAlias) and not (
                        getattr(x, '__origin__', None) in {typing.Generic}),
                bases
            )
        )
        for generic in generics:
            generic_attrs = copy.deepcopy(
                getattr(RoutableMeta.get_type_instance(cls, generic.__name__), GENERIC_ATTRIBUTES)
            )
            for i in range(len(generic_attrs)):
                generic_attrs[i]['type'] = generic.__args__[i]
            setattr(cls, GENERIC_ATTRIBUTES, generic_attrs)

        print(bases)

    def __new__(cls: Type[type], name: str, bases: Tuple[Type[Any]], attrs: Dict[str, Any]) -> 'RoutableMeta':
        RoutableMeta.__new_instance__(cls, name, bases, attrs)
        cls_ = RoutableMeta.get_type_instance(cls, name)
        attrs[CLASS_TYPE] = cls_

        if cls_ is not None:
            methods = RoutableMeta.get_config_endpoints(bases, attrs)
            setattr(cls_, API_METHODS, methods)
            attrs[API_METHODS] = methods

        attrs['routes'] = lambda: RoutableMeta.get_router(RoutableMeta.get_type_instance(cls, name))

        # RoutableMeta.signature(cls, name, bases, attrs)
        if cls_ is not None:
            try:
                prm = cls_.__parameters__[0]
            except Exception as ex:
                pass
            RoutableMeta.init_generic_params(cls_)
            RoutableMeta._init_cbv(cls_)

        return cast(RoutableMeta, type.__new__(cls, name, bases, attrs))


class Routable(metaclass=RoutableMeta):
    """Base class for all classes the want class-based routing.

    This Uses RoutableMeta as a metaclass and various decorators like @get or @post from the decorators module. The
    decorators just mark a method as an endpoint. The RoutableMeta then converts those to a list of desired endpoints in
    the _endpoints class method during class creation. The constructor constructs an APIRouter and adds all the routes
    in the _endpoints to it so they can be added to an app via FastAPI.include_router or similar.
    """
    # _endpoints: List[EndpointDefinition] = []
    VERSION_API = '1.0'
    NAME_MODULE = ''

    BASE_TEMPLATE_PATH = '/{module}/{controller}/v{version}/{user_path}'

    # Тегирование маршрутов API
    TAGS = None
    TAGGING = True



from typing import Dict, Any, Type, Tuple

from class_based_fastapi.utilities import snake_case


def _rest_api_naming(name: str) -> str:
    """Преобразование наименования API

    Args:
        name: Исходное название

    Returns: Преобразованное название
    """
    return snake_case(name).replace('_', '-')


def replace_parameter_template(endpoint: Dict[str, Any], mask_parameter: str, old_value: str, new_value: str) -> None:

    template_path = endpoint.args.template_path
    indx_mask = template_path.find(mask_parameter)
    if indx_mask == -1:
        return

    if indx_mask + len(mask_parameter) < len(template_path):
        mask = f'{template_path[indx_mask - 1]}{old_value}{template_path[indx_mask + len(mask_parameter)]}'
        new_mask = f'{template_path[indx_mask - 1]}{new_value}{template_path[indx_mask + len(mask_parameter)]}'
    else:
        mask = f'{template_path[indx_mask - 1]}{old_value}'
        new_mask = f'{template_path[indx_mask - 1]}{new_value}'

    endpoint.args.path = endpoint.args.path.replace(mask, new_mask)


def _replace_child_name_controller(endpoint: Dict[str, Any],
                                   name: str,
                                   base: Type[Any],
                                   attrs: Dict[str, Any]) -> None:
    """Замена унаследованного названия контроллера

    Args:
        endpoint: Информация о методе
        name: Название текущего класса (контроллера)
        base: Тип унаследованного класса
    """
    old_name = _rest_api_naming(base.__name__)
    new_name = _rest_api_naming(name)

    replace_parameter_template(endpoint, '{controller}', old_name, new_name)


def _replace_child_name_module(endpoint: Dict[str, Any],
                               name: str,
                               base: Type[Any],
                               attrs: Dict[str, Any]) -> None:
    """Замена унаследованного названия контроллера

    Args:
        endpoint: Информация о методе
        name: Название текущего класса (контроллера)
        base: Тип унаследованного класса
    """
    new_name = attrs.get('NAME_MODULE')
    if new_name is None:
        return
    old_name = endpoint.args.name_module

    replace_parameter_template(endpoint, '{module}', old_name, _rest_api_naming(new_name))


def _replace_child_varsion_api(endpoint: Dict[str, Any],
                               name: str,
                               base: Type[Any],
                               attrs: Dict[str, Any]) -> None:
    """Замена унаследованного названия контроллера

    Args:
        endpoint: Информация о методе
        name: Название текущего класса (контроллера)
        base: Тип унаследованного класса

    """
    new_name = attrs.get('VERSION_API')
    if new_name is None:
        return
    old_name = endpoint.args.varsion_api

    replace_parameter_template(endpoint, '{version}', old_name, _rest_api_naming(new_name))


def _processing_child_paths(endpoints: Dict[str, Any],
                            name: str,
                            bases: Tuple[Type[Any]],
                            attrs: Dict[str, Any]) -> Dict[str, Any]:
    for base in bases:
        for endpoint in endpoints.values():

            _replace_child_name_controller(endpoint, name, base, attrs)
            _replace_child_name_module(endpoint, name, base, attrs)
            _replace_child_varsion_api(endpoint, name, base, attrs)

    return endpoints



import copy
import inspect
import re
import types
from functools import partial
from typing import List, TypeVar, Union

from class_based_fastapi.defaults import GENERIC_ATTRIBUTES
from class_based_fastapi.route_args import RouteArgs

_snake_1 = partial(re.compile(r'(.)((?<![^A-Za-z])[A-Z][a-z]+)').sub, r'\1$*-$%\2')
_snake_2 = partial(re.compile(r'([a-z0-9])([A-Z])').sub, r'\1$*-$%\2')


def snake_case(string: str, separator: str = '-') -> str:
    """Преобразование строки в snake case."""
    case = _snake_2(_snake_1(string)).casefold()
    if separator != '_':
        case = case.replace('$*-$%', separator)
    return case


def find_type_from_getneric(name_param: str, generics: List[dict], raise_exception=True) -> Union[dict, None]:
    type_from_generic = list(filter(lambda x: x['name'] == name_param, generics))
    if len(type_from_generic) > 1 or len(type_from_generic) == 0:
        if raise_exception:
            raise Exception(f'Found -> {len(type_from_generic)} <- generic params')
        return None
    return type_from_generic[0]


def _get_response_type(annotation: type, attrs: List[dict]) -> type:
    if isinstance(annotation, TypeVar):
        # type_from_generic = list(filter(lambda x: x['name'] == annotation.__name__, attrs))
        # if len(type_from_generic) > 1 or len(type_from_generic) == 0:
        #     raise Exception
        type_from_generic = find_type_from_getneric(annotation.__name__, attrs)
        if type_from_generic['type'] is not None:
            annotation = type_from_generic['type']
    elif hasattr(annotation, '__origin__'):
        params = []
        for arg in annotation.__args__:
            type_from_generic = find_type_from_getneric(arg.__name__, attrs, raise_exception=False)
            if type_from_generic is not None and type_from_generic['type'] is not None:
                params.append(type_from_generic['type'])
            else:
                params.append(arg)

        annotation = annotation.__origin__[tuple(params)]

    return annotation


def deepcopy_func(f, args: RouteArgs, cls, name=None):
    clone_func = types.FunctionType(
        f.__code__, f.__globals__, name or f.__name__,
        f.__defaults__, f.__closure__
    )
    new_signature = inspect.signature(clone_func)
    old_signature = inspect.signature(f)
    generics = getattr(cls, GENERIC_ATTRIBUTES, [])
    params = dict(old_signature.parameters.items())
    for name_param, param in new_signature.parameters.items():
        if name_param in params:
            annotation = _get_response_type(params[name_param].annotation, generics)
            # if isinstance(annotation, TypeVar):
            #     type_from_generic = list(filter(lambda x: x['name'] == annotation.__name__, generics))
            #     if len(type_from_generic) > 1 or len(type_from_generic) == 0:
            #         raise Exception
            #     annotation = type_from_generic[0]['type']

            params[name_param] = new_signature.parameters[name_param].replace(annotation=annotation)

    new_signature = new_signature.replace(
        parameters=params.values(),
        return_annotation=_get_response_type(old_signature.return_annotation, generics)
    )
    setattr(clone_func, '__signature__', new_signature)

    if args.response_model is None:
        args.response_model = new_signature.return_annotation
    # inspect.signature(f).parameters['model'].annotation = 123
    attrs_names = [k for k in dir(f) if k not in dir(types.FunctionType)]
    for attr in attrs_names:
        setattr(clone_func, attr, copy.deepcopy(getattr(f, attr)))
    return clone_func



from pkg_resources import parse_requirements
from setuptools import setup, find_packages


def load_requirements(fname: str) -> list:
    requirements = []
    with open(fname, 'r', encoding='utf-8') as fp:
        for req in parse_requirements(fp.read()):
            extras = '[{}]'.format(','.join(req.extras)) if req.extras else ''
            requirements.append(
                '{}{}{}'.format(req.name, extras, req.specifier)
            )
    return requirements


setup(
    name='class-based-fastapi',
    version='1.0.2',
    author='XDeepZeroX',
    license='MIT',
    description='Class based routing for FastAPI',
    long_description=open('README.md').read(),
    url='https://github.com/XDeepZeroX/class-based-fastapi',
    platforms='all',
    keywords=["FastAPI", "Class", "Instance", "Routing"],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        "Framework :: FastAPI",
        "Framework :: AsyncIO",
        'Operating System :: OS Independent',
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3 :: Only",
        'Topic :: Internet :: WWW/HTTP',
        "Topic :: Internet :: WWW/HTTP :: Session",
        "Typing :: Typed",
    ],
    python_requires='>=3.8, <4',
    packages=find_packages(exclude=['tests']),
    install_requires=load_requirements('requirements.txt'),
    extras_require={
        'dev': load_requirements('requirements.dev.txt'),
        'test': load_requirements('requirements.test.txt')
    },
    # entry_points={
    #     'console_scripts': [
    #         '{0}-api = {0}.api.__main__:main'.format(module_name),
    #         '{0}-db = {0}.db.__main__:main'.format(module_name)
    #     ]
    # },
    include_package_data=True
)
